# -*- coding: utf-8 -*-
"""
Created on Mon Aug 17 00:02:30 2020

@author: lllan
"""


import numpy as np
import re
import math
from ..utils.constants import *
from ..utils.elements import ELEMENTS, elements_inversed
from ..comms import printer
from ..tests import symmetry
from ..core import ElasticProperties
from ..core import Structure

def mag(v):
    return (v[0]*v[0]+v[1]*v[1]+v[2]*v[2])**0.5
def vol(a,b,c):
    return a[0]*(b[1]*c[2] - b[2]*c[1]) - a[1]*(b[0]*c[2] - b[2]*c[0]) + a[2]*(b[0]*c[1] - b[1]*c[0])

class QE_thermo_pw_Parser:

    """
    This class contains methods to parse the OUTCAR file.
    """

    def __init__(self, outfile = 'scf.out', infile = 'scf.in'):


        self.outfile = outfile
        self.infile  = infile

        
        self.elastic_tensor = None
        self.compaliance_tensor = None
        self.structure = None
        self.density = None
                
        self.outText = None
        self.inText  = None
        
        self._parse_files()


        return

    def _parse_files(self):
        """
        This function reads outcar and initializes the class variables

        """


        rf = open(self.outfile)
        self.outText = rf.read()
        rf.close()
        dataOut= self.outText
        
        rf = open(self.infile)
        self.inText = rf.read()
        rf.close()
        dataIn= self.inText
        
        nspecies = int(re.findall("\s*ntyp\s*=\s*(\d)",dataIn)[0])

        

        
        nions = int(re.findall("\s*nat\s*=\s*(\d)",dataIn)[0])
        raw_ions = re.findall("ATOMIC_POSITIONS.*\n" + nions * "(.*)\n",dataIn)[0]
        
        composition = {}
        species_list = []
        mass = []
        raw_species = re.findall("ATOMIC_SPECIES.*\n" + nspecies * "(.*).*\n",dataIn)[0]
        
        if(nspecies == 1):
            composition[raw_species.split()[0]] = {'numAtom':0, 'mass':raw_species.split()[1]}
            species_list.append(raw_species.split()[0])
            
        else:
            for nspec in range(nspecies):
                species_list.append(raw_species[nspec].split()[0])
                composition[raw_species.split()[0]] = {'numAtom':0, 'mass':raw_species.split()[1]}
                
        if(nions == 1):
            composition[raw_species.split()[0]]['numAtom'] = 1
        else:
            for ions in range(nions):
                for species in range(len(species_list)):
                    if(raw_ions[ions].split()[0] == species_list[species]):
                        composition[raw_ions[ions].split()[0]]['numAtom'] += 1

        positions = np.array(
            [
                x.strip().split()[1:]
                for x in raw_ions
            ]
        ).astype(float)
        
        
        iontype = [int(composition[x]['numAtom']) for x in list(composition.keys())]
       
        ions = [x.strip().split()[0] for x in raw_ions]
        mass  = [float(composition[x]['mass']) for x in ions]
        
        species = list(composition.keys())
        
        ibrav = int(re.findall("\s*ibrav\s*=\s*(\d*)",dataIn)[0])
        
        if (ibrav == 0):
            num_latticeVectors = 3
            celldm1 = float(re.findall("\s*celldm\(1\)\s*=\s*([\.\d]*)",data)[0])
            raw_cell = re.findall("CELL_PARAMETERS.*\n" + num_latticeVectors * "(.*)\n",data)[0]
            lattice = np.array([[float(component) for component in row.split()] for row in raw_cell])
            lattice = lattice*celldm1
            a = mag(lattice[0,:])
            b = mag(lattice[1,:])
            c = mag(lattice[2,:])
        if (ibrav == 1 or ibrav == 2 or ibrav == 3):
            "cubic systems"
            celldm1 = float(re.findall("\s*celldm\(1\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm2 = 1. 
            celldm3 = 1. 
            celldm4 = 0.
            celldm5 = 0.
            celldm6 = 0.
            celldm2 = 1.
            
            if(ibrav == 1):
                lattice = np.array([[celldm1,0,0],
                                    [0,celldm1,0],
                                    [0,0,celldm1]])
                a = mag(lattice[0,:])
                b = mag(lattice[1,:])
                c = mag(lattice[2,:])
            elif(ibrav == 2):
                lattice = np.array([[-celldm1/2,0,celldm1/2],
                                    [0,celldm1/2,celldm1/2],
                                    [-celldm1/2,celldm1/2,0]])
                a = mag(lattice[0,:])
                b = mag(lattice[1,:])
                c = mag(lattice[2,:])
            elif(ibrav == 3):
                lattice = np.array([[celldm1/2,0,celldm1/2],
                                    [-celldm1/2,celldm1/2,celldm1/2],
                                    [-celldm1/2,-celldm1/2,celldm1/2]])
                a = mag(lattice[0,:])
                b = mag(lattice[1,:])
                c = mag(lattice[2,:])
        if (ibrav == 4):
            "Hexagonal system's must have celldm(1) and celldm(3)"
            celldm1 = float(re.findall("\s*celldm\(1\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm2 = 1. 
            celldm3 = float(re.findall("\s*celldm\(3\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm4 = 0.
            celldm5 = 0.
            celldm6 =-.5
            
            lattice = np.array([[celldm1,0,0],
                                [-celldm1/2,celldm1*sqrt(3)/2,0],
                                [0,0,celldm1*celldm3]])
            a = mag(lattice[0,:])
            b = mag(lattice[1,:])
            c = mag(lattice[2,:])
            
        if (ibrav == 5):
            "Trigonal or Rhombohedragonal structures must have celldm(1) and celldm(4)"
            celldm1 = float(re.findall("\s*celldm\(1\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm2 = 1. 
            celldm3 = 1.
            celldm4 = float(re.findall("\s*celldm\(4\)\s*=\s*([\.\d]*)",dataIn)[0])
            if (celldm4 < -1 or celldm4 > 1.):
                sys.exit('\n.... Oops ERROR: "celldm(4)" is WRONG !?!?!?    \n')
            celldm5 = celldm4 
            celldm6 = celldm4
            
            tx = ((1-celldm4)/2)**0.5
            ty = ((1-celldm4)/6)**0.5
            tz = ((1+2*celldm4)/3)**0.5
            
            lattice = np.array([[celldm1*tx,-celldm1*ty,celldm*tz],
                                [0,2*celldm1*ty,tz],
                                [-celldm1*tx,-celldm1*ty,celldm1*tz]])
            a = mag(lattice[0,:])
            b = mag(lattice[1,:])
            c = mag(lattice[2,:])
            
            
            
        if (ibrav == 6 or ibrav == 7):
            "Tetragonal structures must have celldm(1) and celldm(3)"
            celldm1 = float(re.findall("\s*celldm\(1\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm2 = 1.
            celldm3 = float(re.findall("\s*celldm\(3\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm4 = .0 
            celldm5 = .0
            celldm6 = .0
            
            if(ibrav == 6):
                lattice = np.array([[celldm1,0,0],
                                    [0,celldm1,0],
                                    [0,0,celldm1*celldm3]])
                a = mag(lattice[0,:])
                b = mag(lattice[1,:])
                c = mag(lattice[2,:])

            elif(ibrav == 7):
                lattice = np.array([[celldm1/2,-celldm1/2,celldm1*celldm3/2],
                                    [celldm1/2,celldm1/2,celldm1*celldm3/2],
                                    [-celldm1/2,-celldm1/2,celldm1*celldm3/2]])
                a = mag(lattice[0,:])
                b = mag(lattice[1,:])
                c = mag(lattice[2,:])
        
        if (ibrav == 8 or ibrav == 9 or ibrav == 10 or ibrav == 11):
            "Orthorhombic structures must have celldm(1), celldm(2), and celldm(3)"
            celldm1 = float(re.findall("\s*celldm\(1\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm2 = float(re.findall("\s*celldm\(2\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm3 = float(re.findall("\s*celldm\(3\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm4 = .0
            celldm5 = .0
            celldm6 = .0
            
            if(ibrav == 8):
                lattice = np.array([[celldm1,0,0],
                                    [0,celldm1*celldm2,0],
                                    [0,0,celldm1*celldm3]])
                a = mag(lattice[0,:])
                b = mag(lattice[1,:])
                c = mag(lattice[2,:])
            elif(ibrav == 9):
                lattice = np.array([[celldm1/2,celldm1*celldm2/2,0],
                                    [-celldm1/2,celldm1*celldm2/2,0],
                                    [0,0,celldm1*celldm3]])
                a = mag(lattice[0,:])
                b = mag(lattice[1,:])
                c = mag(lattice[2,:])
            elif(ibrav == 10):
                lattice = np.array([[celldm1/2,0,celldm1*celldm3/2],
                                    [celldm1/2,celldm1*celldm2/2,0],
                                    [0,celldm1*celldm2/2,celldm1*celldm3/2]])
                a = mag(lattice[0,:])
                b = mag(lattice[1,:])
                c = mag(lattice[2,:])
            elif(ibrav == 11):
                lattice = np.array([[celldm1/2,celldm1*celldm2/2,celldm1*celldm3/2],
                                    [-celldm1/2,celldm1*celldm2/2,celldm1*celldm3/2],
                                    [-celldm1/2,-celldm1*celldm2/2,celldm1*celldm3/2]])
                a = mag(lattice[0,:])
                b = mag(lattice[1,:])
                c = mag(lattice[2,:])
            
        if (ibrav == 12 or ibrav == 13):
            "Monoclinic structures must have celldm(1), celldm(2), and celldm(3),celldm(4)"
            
            
            celldm1 = float(re.findall("\s*celldm\(1\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm2 = float(re.findall("\s*celldm\(2\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm3 = float(re.findall("\s*celldm\(3\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm4 = float(re.findall("\s*celldm\(4\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm5 = .0
            celldm6 = .0
            
            if(ibrav == 12):
                lattice = np.array([[celldm1/2,0,0],
                                    [celldm1*celldm2*celldm4,celldm1*celldm2*celldm4,0],
                                    [0,0,celldm1*celldm3]])
                a = mag(lattice[0,:])
                b = mag(lattice[1,:])
                c = mag(lattice[2,:])
            elif(ibrav == 13):
                lattice = np.array([[celldm1/2,0,-celldm1*celldm3/2],
                                    [celldm1*celldm2*celldm4,celldm1*celldm2*celldm4,0],
                                    [celldm1/2,0,celldm1*celldm3/2]])
                a = mag(lattice[0,:])
                b = mag(lattice[1,:])
                c = mag(lattice[2,:])

        if (ibrav == 14):
            "Triclinic structure must have celldm(1), celldm(2), and celldm(3),celldm(4),celldm(5), and celldm(6)"
            celldm1 = float(re.findall("\s*celldm\(1\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm2 = float(re.findall("\s*celldm\(2\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm3 = float(re.findall("\s*celldm\(3\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm4 = float(re.findall("\s*celldm\(4\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm5 = float(re.findall("\s*celldm\(5\)\s*=\s*([\.\d]*)",dataIn)[0])
            celldm6 = float(re.findall("\s*celldm\(6\)\s*=\s*([\.\d]*)",dataIn)[0])
            
            alpha = math.acos(celldm4)
            beta  = math.acos(celldm5)
            gamma = math.acos(celldm6)
           
            lattice = np.array([[celldm1,0,0],
                                [celldm1*celldm2*math.cos(gamma), celldm1*celldm2*math.sin(gamma), 0],
                                [celldm1*celldm3*math.cos(beta),  celldm1*celldm3*(math.cos(alpha)-math.cos(beta)*math.cos(gamma))/math.sin(gamma),
            celldm1*celldm3*math.sqrt( 1 + 2*math.cos(alpha)*math.cos(beta)*math.cos(gamma)
                     - math.cos(alpha)**2-math.cos(beta)**2-math.cos(gamma)**2 )/math.sin(gamma)]])
            
            a = mag(lattice[0,:])
            b = mag(lattice[1,:])
            c = mag(lattice[2,:])
            
        
        volume = vol(lattice[0,:],lattice[1,:],lattice[2,:])
        
        # Convert to a.u. (bohr) to Angtroms
        lattice = lattice*(0.529177249)   
        self.lattice_constant = [a*0.529177249  ,b*0.529177249  ,c*0.529177249  ]         


        # cell parameters
        A = float(self.lattice_constant[0])
        B = float(self.lattice_constant[1])
        C = float(self.lattice_constant[2])

        print(
            "Lattice parameters (in Angs.): a = %10.5f      b = %10.5f     c = %10.5f"
            % (A, B, C)
        )

        atomic_numbers = np.zeros(nions, dtype=np.int32)
        k = 0
        for i in range(len(iontype)):
            for j in range(iontype[i]):
                atomic_numbers[k] = ELEMENTS[species[i]]
                k = k + 1
        

        print("Atomic numbers")
        print(atomic_numbers)
        atomic_numbers = atomic_numbers
        symbols = [elements_inversed[x] for x in atomic_numbers]
        cell = (lattice, positions, atomic_numbers)

        print("Mass of atoms (in g/mol units): ")
        print((np.array(mass)))
        print("Number of atoms: %d" % np.sum(nions))

        total_mass = 0.0
        for i in range(len(ions)):
            total_mass = total_mass +  mass[i]
        print("Total mass (in g/mol): %10.4f " % total_mass)
        volAng= volume*0.529177249**3
        print("Volume of the cell (in Ang^3 units): %10.4f " % volAng)

        # converting the units
        volume *= (5.29177249**-11)**3  # from atomic units (bohr)  to meters
        total_mass *= 1.0e-3  # from gram to kg
        density = total_mass / (volume * N_avogadro)
        self.density = density 
        print("\nDensity (in kg/m^3 units ): %10.5f" % density)


        num_rows = 6
        raw_stiffness = re.findall("\s*Elastic\sconstants.*\n.*\n" + num_rows * "\s*(.*)\n",dataOut)[0]
        c = np.array([[float(x)/10.0 for x in row.split()[1:]] for row in raw_stiffness])

    
        print("\nPrinting Cij matrix as read from outfile (GPa)\n")
        self.elastic_tensor = c.copy()
        np.set_printoptions(precision=3, suppress=True)
        printer.printMatrix(self.elastic_tensor)
        
        self.compaliance_tensor = np.linalg.inv(self.elastic_tensor)


        print(
            (
                "\n Checking if the modified matrix CNEW is symmetric: i.e. Cij = Cji:  %10s"
                % symmetry.check_symmetric(self.elastic_tensor)
            )
        )
        self.structure = Structure(symbols, positions, lattice)

        return



import numpy as np
import re
from ..utils.constants import *
from ..utils.elements import ELEMENTS, elements_inversed

from ..comms import printer
from ..tests import symmetry

from ..core import ElasticProperties
from ..core import Structure


class VaspOutcar:

    """
    This class contains methods to parse the OUTCAR file.
    """

    def __init__(self, infile="OUTCAR", adjust_pressure=True, verbose=True):

        self.infile = infile
        self.adjust_pressure = adjust_pressure

        self.elastic_tensor = None
        self.compaliance_tensor = None
        self.structure = None
        self.density = None
        self.text = None

        self.mass = None
        self.volume = None
        self.nions = None

        self.lattice_type = None
        self.lattice = None
        self.positions = None
        self.iontype = None
        self.species = None
        self.lattice_constant = None
        self.pressure = None
        self.total_mass = None

        self._parse_outcar(verbose)
        if verbose:
            print(self)
        # self.elastic_properties = ElasticProperties(
        #     self.elastic_tensor, self.structure)#, self.crystal_type)

        return

    def _parse_outcar(self, verbose):
        """
        This function reads outcar and initializes the class variables

        """
        if verbose:
            print("Parsing %s" % self.infile)
        rf = open(self.infile)
        self.text = rf.read()
        rf.close()
        data = self.text
        self.mass = np.array(
            [float(x)
             for x in re.findall("POMASS\s*=\s*([0-9.]*);\s*ZVAL", data)]
        )
        self.volume = float(re.findall(
            "volume of cell\s*:\s*([0-9.]*)", data)[-1])
        self.nions = int(re.findall("NIONS\s*=\s*([0-9]*)", data)[0])
        self.natoms = np.array(
            [
                int(x)
                for x in re.findall("ions per type\s*=\s*([\s0-9]*)", data)[0].split()
            ]
        )

        # Sometimes LATTYP is not present.
        # Only parse if available.
        self.lattice_type = re.findall("LATTYP.*", data)
        if self.lattice_type:
            lattyp = self.lattice_type[-1]
        self.lattice = np.array(
            [
                x.split()[:3]
                for x in re.findall(
                    "direct lattice vectors.*\n([-+0-9.\s\n]*)length of vectors", data
                )[-1].split("\n")[:3]
            ]
        ).astype(float)
        self.positions = np.array(
            [
                x.split()[:3]
                for x in re.findall(
                    "position of ions in fractional coordinates.*\n([-+0-9.\s\n]*)",
                    data,
                )[-1].split("\n")[:self.nions]
            ]
        ).astype(float)
        self.iontype = [
            int(x)
            for x in re.findall("ions per type\s*=\s*([0-9\s]*)", data)[0].split()
        ]
        self.species = re.findall("VRHFIN\s*=([a-zA-Z\s]*):", data)
        self.lattice_constant = [
            float(x)
            for x in re.findall("length of vectors.*\n([0-9.\s]*)", data)[-1].split()[
                :3
            ]
        ]

        # external pressure in kB -> GPa
        self.pressure = float(
            re.findall(r"external\s*pressure\s=\s*([-0-9.]*)\s*kB", data)[-1]
        )
        self.pressure = self.pressure / 10

        atomic_numbers = np.zeros(self.nions, dtype=np.int32)
        k = 0
        for i in range(len(self.iontype)):
            for j in range(self.iontype[i]):
                atomic_numbers[k] = ELEMENTS[self.species[i].strip()]
                k = k + 1

        self.atomic_numbers = atomic_numbers
        self.symbols = [elements_inversed[x] for x in atomic_numbers]

        self.total_mass = 0.0
        for i in range(len(self.natoms)):
            self.total_mass += self.natoms[i] * self.mass[i]

        # converting the units
        self.volume *= 1.0e-30  # from Angstrom to meters
        self.total_mass *= 1.0e-3  # from gram to kg
        self.density = self.total_mass / (self.volume * N_avogadro)
        self.density = self.density

        c = np.matrix(
            [
                x.split()[1:]
                for x in re.findall(
                    "TOTAL ELASTIC MODULI \(kBar\).*\n.*\n.*\n([XYZ0-9.\s-]*)\n\s*-",
                    data,
                )[0].split("\n")
            ][:6]
        ).astype(float)

        # compaliance_tensor = c.I
        if verbose:
            print("\nPrinting Elastic Tensor, Cij as read from OUTCAR\n")

            np.set_printoptions(precision=4, suppress=True)
            printer.print_matrix(c)

        self.elastic_tensor = c.copy()

        # Redefining the Cij matrix into the correct Voigt notation since VASP's OUTCAR has a different order
        # In VASP: Columns and rows are listed as: 1, 2, 3, 6, 4, 5
        # In this format OUTCAR's C44 values would be actually C66, C55 would be C44, and C66 would be C55.
        # OUTCAR follows the below order:
        # [C11 C12 C13 C16 C14 C15]
        # [C21 C22 C23 C26 C24 C25]
        # [C31 C32 C33 C36 C34 C35]
        # [C61 C62 C63 C66 C64 C65]
        # [C41 C42 C43 C46 C44 C45]
        # [C51 C52 C53 C56 C54 C55]

        for j in range(0, 6):
            self.elastic_tensor[3, j] = c[4, j]
            self.elastic_tensor[4, j] = c[5, j]
            self.elastic_tensor[5, j] = c[3, j]

        ctemp = self.elastic_tensor.copy()

        for i in range(0, 6):
            self.elastic_tensor[i, 3] = self.elastic_tensor[i, 4]
            self.elastic_tensor[i, 4] = self.elastic_tensor[i, 5]
            self.elastic_tensor[i, 5] = ctemp[i, 3]

        # Change the units of Cij from kBar to GPa
        self.elastic_tensor /= 10.0

        # Print elastic tensor in GPa units
        if verbose:
            print(
                "\n \n printing Elastic Tensor: corrected order (in GPa units)... \n For example- to generate input for the ELATE code [https://github.com/fxcoudert/elate] \n \n"
            )
            np.set_printoptions(precision=3, suppress=True)
            printer.print_matrix(self.elastic_tensor)

        if self.adjust_pressure and self.pressure != 0.0:
            # In VASP the pressure needs to be adjusted in the elastic tensor.
            # Subtract P in the diagonal and add P in C12, C21, C13, C31, C23, C32.
            if verbose:
                print(
                    "\nAdjusted for pressure since non-zero hydrostatic pressure exists.")
                print("External Pressure : %10.5f GPa" % self.pressure)
                print(
                    "Set flag adjust_pressure=False to disable. -ap 0 in stand-alone mode."
                )

            for i in range(0, 6):
                self.elastic_tensor[i,
                                    i] = self.elastic_tensor[i, i] - self.pressure
            self.elastic_tensor[0,
                                1] = self.elastic_tensor[0, 1] + self.pressure
            self.elastic_tensor[1,
                                0] = self.elastic_tensor[1, 0] + self.pressure
            self.elastic_tensor[0,
                                2] = self.elastic_tensor[0, 2] + self.pressure
            self.elastic_tensor[2,
                                0] = self.elastic_tensor[2, 0] + self.pressure
            self.elastic_tensor[1,
                                2] = self.elastic_tensor[1, 2] + self.pressure
            self.elastic_tensor[2,
                                1] = self.elastic_tensor[2, 1] + self.pressure
            if verbose:
                print("\nPressure adjusted Elastic Tensor in GPa units:\n")
                np.set_printoptions(precision=3, suppress=True)
                printer.print_matrix(self.elastic_tensor)

        self.compaliance_tensor = self.elastic_tensor.I

        if verbose:
            print(
                (
                    "\n Checking if the modified Elastic Tensor is symmetric: i.e. Cij = Cji:  %10s"
                    % symmetry.check_symmetric(self.elastic_tensor)
                )
            )
        self.structure = Structure(self.symbols, self.positions, self.lattice)

        return

    def __str__(self):
        ret = ""
        # cell parameters
        A = float(self.lattice_constant[0])
        B = float(self.lattice_constant[1])
        C = float(self.lattice_constant[2])

        ret += (
            "Lattice parameters (in Angs.): a = %10.5f      b = %10.5f     c = %10.5f\n"
            % (A, B, C)
        )

        ret += ("Mass of atoms (in g/mol units): \n" +
                "  ".join(["%s " for x in np.array(self.mass)]) % tuple(self.mass))
        ret += ("Number of atoms: %d\n" % sum(self.natoms))
        ret += ("Total mass (in g/mol): %10.4f \n" % self.total_mass)
        ret += ("Volume of the cell (in Ang^3 units): %10.4f \n" % self.volume)
        ret += ("\nDensity (in kg/m^3 units ): %10.5f" % self.density)
        ret += ("External Pressure (in GPa units ): %10.5f" % self.pressure)
        return ret



#!/usr/bin/env python

"""This file contains physical constants."""

# Avogadro number
N_avogadro = 6.022140857e23
# Planck's constant
h_Planck = 6.62607004e-34
# Boltzmann constant J/K
kB = 1.38064852e-23
# Atomic mass unit. kg
amu = 1.66053886e-27
